import pytest

from attrs import evolve
from mantelo.connection import (
    Token,
    AuthenticationException,
    ClientCredentialsConnection,
    UsernamePasswordConnection,
)
from datetime import datetime, timedelta, timezone
import requests


def test_token_from_dict():
    now = datetime.now(timezone.utc)
    one_sec = timedelta(seconds=1)

    token = Token.from_dict(
        data={
            "access_token": "tok",
            "expires_in": 3600,
            "refresh_token": "rtok",
            "refresh_expires_in": 6000,
            "scope": "openid",
            # ignored
            "foo": "bar",
        }
    )

    assert token.access_token == "tok"
    assert token.refresh_token == "rtok"
    assert token.scope == "openid"
    assert token.token_type is None
    assert now - one_sec < token.created_at < now + one_sec


def test_refresh_without_expire():
    with pytest.raises(TypeError, match="Missing refresh_expires_in."):
        Token(
            access_token="<any>",
            expires_in=123,
            refresh_token="<set>",
        )


@pytest.mark.parametrize(
    "data",
    [
        {},
        {"access_token": "set"},
        {"access_token": "set", "refresh_token": "set"},
    ],
)
def test_token_from_dict_fails(data):
    with pytest.raises(TypeError):
        Token.from_dict(data=data)


def test_expires_at():
    now = datetime.now(timezone.utc)
    token = Token(
        access_token="<any>",
        expires_in=3600,
        refresh_token="<any>",
        refresh_expires_in=3600,
        created_at=now - timedelta(seconds=3600),
    )

    assert token.expires_at == now
    assert token.refresh_expires_at == now

    token = evolve(token, refresh_token=None)
    assert token.refresh_expires_at is None


@pytest.mark.parametrize(
    ("ok", "token", "now_delta"),
    [
        (False, None, 0),
        (False, "<rt>", 3600),
        (False, "<rt>", 3598),
        (True, "<rt>", 3597),
        (True, "<rt>", 0),
    ],
)
def test_token_has_refresh_token(ok, token, now_delta):
    now = datetime.now(timezone.utc)

    token = Token(
        access_token="<any>",
        expires_in=3600,
        refresh_token=token,
        refresh_expires_in=3600,
        created_at=now,
    )
    res = token.has_refresh_token(
        _now=lambda: now + timedelta(seconds=now_delta)
    )

    assert res == ok


def test_userpasswordconnection_init():
    conn = UsernamePasswordConnection(
        server_url="https://kc.test",
        realm_name="test",
        client_id="foo-client",
        username="u",
        password="p",
    )
    assert (
        conn.auth_url
        == "https://kc.test/realms/test/protocol/openid-connect/token"
    )
    assert conn._token_exchange_data() == {
        "scope": "openid",
        "grant_type": "password",
        "client_id": "foo-client",
        "username": "u",
        "password": "p",
    }
    assert isinstance(conn.session, requests.Session)
    assert conn.refresh_timeout == timedelta(seconds=30)


def test_ClientCredentialsconnection_init():
    conn = ClientCredentialsConnection(
        server_url="https://kc.test",
        realm_name="test",
        client_id="foo-client",
        client_secret="s3cr3t",
    )
    assert (
        conn.auth_url
        == "https://kc.test/realms/test/protocol/openid-connect/token"
    )
    assert conn._token_exchange_data() == {
        "scope": "openid",
        "grant_type": "client_credentials",
        "client_id": "foo-client",
        "client_secret": "s3cr3t",
    }
    assert isinstance(conn.session, requests.Session)
    assert conn.refresh_timeout == timedelta(seconds=30)


def test_openidconnection_default_values():
    args = dict(
        server_url="https://kc.test",
        realm_name="test",
        client_id="foo-client",
        client_secret="s3cr3t",
    )

    # Defaults
    conn = ClientCredentialsConnection(**args)
    assert isinstance(conn.session, requests.Session)
    assert conn.refresh_timeout == timedelta(seconds=30)

    # Override defaults
    session = requests.Session()
    timeout = timedelta(seconds=120)
    conn = ClientCredentialsConnection(
        **args,
        session=session,
        refresh_timeout=timeout,
    )
    assert conn.session == session
    assert conn.refresh_timeout == timeout

    # None values fallback to the default values
    conn = ClientCredentialsConnection(
        **args,
        session=None,
        refresh_timeout=None,
    )
    assert isinstance(conn.session, requests.Session)
    assert conn.refresh_timeout == timedelta(seconds=30)


@pytest.mark.integration
def test_openid_token(openid_connection_password):
    assert not openid_connection_password._token

    # first fetch
    access_token = openid_connection_password.token()
    assert access_token

    # second fetch: same
    assert openid_connection_password.token() == access_token

    # token expired: use refresh token
    openid_connection_password._token = evolve(
        openid_connection_password._token,
        created_at=datetime.now(timezone.utc) - timedelta(days=1),
    )
    assert openid_connection_password.token() != access_token

    # token expired and no refresh still works
    access_token = openid_connection_password.token()
    openid_connection_password._token = evolve(
        openid_connection_password._token,
        refresh_token=None,
        created_at=datetime.now(timezone.utc) - timedelta(days=1),
    )
    assert openid_connection_password.token() != access_token
    assert openid_connection_password._token.expires_at > datetime.now(
        timezone.utc
    )


@pytest.mark.integration
def test_openid_refresh_token(openid_connection_password):
    assert not openid_connection_password._token

    openid_connection_password._fetch_token()
    token = openid_connection_password._token
    assert token
    assert token.has_refresh_token()

    openid_connection_password.password = "invalid"
    openid_connection_password._fetch_token()
    assert openid_connection_password._token
    assert openid_connection_password._token.expires_at > token.expires_at


@pytest.mark.integration
def test_openid_token_fails(openid_connection_password):
    assert not openid_connection_password._token

    openid_connection_password.password = "wrong"
    with pytest.raises(AuthenticationException) as excinfo:
        openid_connection_password.token()

    assert excinfo.value.error == "invalid_grant"
    assert excinfo.value.error_description == "Invalid user credentials"

    openid_connection_password.username = None
    with pytest.raises(AuthenticationException) as excinfo:
        openid_connection_password.token()

    assert excinfo.value.error == "invalid_request"
    assert excinfo.value.error_description == "Missing parameter: username"



import requests
from attrs import field, frozen
from slumber.exceptions import SlumberHttpBaseException


@frozen
class AuthenticationException(Exception):
    error: str
    error_description: str


@frozen
class HttpException(Exception):
    status_code: int
    json: str
    url: str
    response: requests.Response = field(repr=False)

    @classmethod
    def from_slumber_exception(cls, ex: SlumberHttpBaseException):
        assert hasattr(ex, "response")
        return cls(
            url=ex.response.request.url,
            status_code=ex.response.status_code,
            json=ex.response.json(),
            response=ex.response,
        )



from abc import ABC, abstractmethod
from collections.abc import Callable
from datetime import datetime, timedelta, timezone
from logging import getLogger

import requests
from attrs import Factory, define, field, frozen

from .exceptions import AuthenticationException


logger = getLogger(__name__)

_NO_AUTH = lambda r: r  # noqa: E731


# TODO: those converters are ugly, but due to a bug in mypy
# (see https://github.com/python/mypy/issues/6535),
# it is not possible to use the `default_if_none` converter
# for now.
# Proper field arguments to use once fixed:
#   converter=default_if_none(timedelta(seconds=30)),
#   converter=default_if_none(Factory(requests.Session)),
#
def _session_if_none_converter(value) -> requests.Session:
    if value is None:
        return requests.Session()
    assert isinstance(value, requests.Session)
    return value


def _timedelta_if_none_converter(value: timedelta | None) -> timedelta:
    if value is None:
        return timedelta(seconds=30)
    return value


def utcnow():
    return datetime.now(timezone.utc)


class Connection(ABC):
    @abstractmethod
    def token(self) -> str:
        pass  # NOCOV


@frozen
class Token:
    access_token: str
    expires_in: int
    scope: str | None = None
    token_type: str | None = None
    refresh_expires_in: int | None = None
    refresh_token: str | None = None
    created_at: datetime = field(default=Factory(utcnow))

    def __attrs_post_init__(self):
        if self.refresh_token and not self.refresh_expires_in:
            raise TypeError("Missing refresh_expires_in.")

    @property
    def expires_at(self) -> datetime:
        return self.created_at + timedelta(seconds=self.expires_in)

    @property
    def refresh_expires_at(self) -> datetime | None:
        if not self.refresh_token:
            return None
        assert self.refresh_expires_in
        return self.created_at + timedelta(seconds=self.refresh_expires_in)

    def has_refresh_token(self, _now: Callable[[], datetime] = utcnow) -> bool:
        if (exp := self.refresh_expires_at) and _now() < (
            exp - timedelta(seconds=2)
        ):
            return True
        return False

    @classmethod
    def from_dict(cls, data: dict, now: datetime | None = None) -> "Token":
        assert hasattr(cls, "__slots__")
        if now:
            data["created_at"] = now
        return cls(**{k: v for k, v in data.items() if k in cls.__slots__})


@define
class OpenidConnection(Connection, ABC):
    server_url: str
    realm_name: str
    client_id: str

    session: requests.Session = field(
        default=Factory(requests.Session),
        converter=_session_if_none_converter,
        kw_only=True,
    )
    refresh_timeout: timedelta = field(
        default=timedelta(seconds=30),
        converter=_timedelta_if_none_converter,
        kw_only=True,
    )

    _token: Token | None = field(
        init=False, repr=False, eq=False, default=None
    )

    @property
    def auth_url(self) -> str:
        return f"{self.server_url}/realms/{self.realm_name}/protocol/openid-connect/token"

    @abstractmethod
    def _token_exchange_data(self) -> dict:
        pass  # NOCOV

    def _fetch_token(self) -> None:
        now = utcnow()

        if self._token and self._token.has_refresh_token():
            logger.debug("Refreshing token")
            data = {
                "grant_type": "refresh_token",
                "client_id": self.client_id,
                "refresh_token": self._token.refresh_token,
            }
        else:
            logger.debug("Fetching token")
            data = self._token_exchange_data()

        # Ensure the call does not use authentication,
        # to avoid recursion errors.
        resp = self.session.post(self.auth_url, data=data, auth=_NO_AUTH)
        if resp.status_code == 401:
            raise AuthenticationException(**resp.json())
        resp.raise_for_status()
        self._token = Token.from_dict(resp.json(), now=now)
        logger.debug(
            "Token valid for %s, refresh token valid for %s",
            self._token.expires_in,
            self._token.refresh_expires_in,
        )

    def token(self, _now: Callable[[], datetime] = utcnow) -> str:
        if not self._token or _now() > (
            self._token.expires_at - self.refresh_timeout
        ):
            self._fetch_token()

        assert self._token
        return self._token.access_token


@define
class UsernamePasswordConnection(OpenidConnection):
    username: str
    password: str

    def _token_exchange_data(self) -> dict:
        return {
            "scope": "openid",
            "grant_type": "password",
            "client_id": self.client_id,
            "username": self.username,
            "password": self.password,
        }


@define
class ClientCredentialsConnection(OpenidConnection):
    client_secret: str

    def _token_exchange_data(self) -> dict:
        return {
            "scope": "openid",
            "grant_type": "client_credentials",
            "client_id": self.client_id,
            "client_secret": self.client_secret,
        }



from mantelo.connection import (
    UsernamePasswordConnection,
    ClientCredentialsConnection,
)
from . import constants
import pytest


@pytest.fixture()
def openid_connection_password():
    return UsernamePasswordConnection(
        server_url=constants.TEST_SERVER_URL,
        realm_name=constants.TEST_REALM,
        client_id=constants.ADMIN_CLI_CLIENT,
        username=constants.TEST_USER,
        password=constants.TEST_PASSWORD,
    )


@pytest.fixture()
def openid_connection_sa():
    return ClientCredentialsConnection(
        server_url=constants.TEST_SERVER_URL,
        realm_name=constants.TEST_REALM,
        client_id=constants.TEST_CLIENT_ID,
        client_secret=constants.TEST_CLIENT_SECRET,
    )


@pytest.fixture()
def openid_connection_admin():
    return UsernamePasswordConnection(
        server_url=constants.TEST_SERVER_URL,
        realm_name=constants.MASTER_REALM,
        client_id=constants.ADMIN_CLI_CLIENT,
        username=constants.TEST_MASTER_USER,
        password=constants.TEST_MASTER_PASSWORD,
    )



from collections.abc import Callable

import requests
import slumber
from attrs import define
from slumber.exceptions import SlumberHttpBaseException

from .connection import (
    ClientCredentialsConnection,
    OpenidConnection,
    UsernamePasswordConnection,
)
from .exceptions import HttpException


class HypenatedResourceMixin(slumber.ResourceAttributesMixin):
    def __getattr__(self, item):
        if item.startswith("_"):
            raise AttributeError(item)
        return super().__getattr__(item.replace("_", "-"))


class HypenatedResource(HypenatedResourceMixin, slumber.Resource):
    def _request(self, *args, **kwargs):
        try:
            return slumber.Resource._request(self, *args, **kwargs)
        except SlumberHttpBaseException as ex:
            raise HttpException.from_slumber_exception(ex) from None


@define
class BearerAuth(requests.auth.AuthBase):
    token_getter: Callable[[], str]

    def __call__(self, r):
        r.headers["Authorization"] = f"Bearer {self.token_getter()}"
        return r


class KeycloakAdmin(HypenatedResourceMixin, slumber.API):
    resource_class = HypenatedResource

    def __init__(
        self,
        server_url: str,
        realm_name: str,
        auth: requests.auth.AuthBase,
        session: requests.Session | None = None,
    ):
        super().__init__(
            base_url=f"{server_url}/admin/realms/{realm_name}",
            auth=auth,
            session=session,
            append_slash=False,
        )

    @property
    def session(self):
        return self._store["session"]

    @property
    def base_url(self):
        return self._store["base_url"]

    @property
    def realm_name(self):
        return self._store["base_url"].split("/realms/")[1]

    @realm_name.setter
    def realm_name(self, realm_name):
        base_url = self._store["base_url"].split("/realms/")[0]
        self._store["base_url"] = f"{base_url}/realms/{realm_name}"

    @classmethod
    def create(
        cls,
        connection: OpenidConnection,
        realm_name: str | None = None,
    ):
        return cls(
            connection.server_url,
            realm_name or connection.realm_name,
            BearerAuth(connection.token),
            session=connection.session,
        )

    @classmethod
    def from_client_credentials(
        cls,
        server_url: str,
        realm_name: str,
        client_id: str,
        client_secret: str,
        authentication_realm_name: str | None = None,
        session: requests.Session | None = None,
    ):
        openid_connection = ClientCredentialsConnection(
            server_url=server_url,
            realm_name=authentication_realm_name or realm_name,
            client_id=client_id,
            client_secret=client_secret,
            session=session,
        )
        return cls.create(
            openid_connection,
            realm_name=realm_name,
        )

    @classmethod
    def from_username_password(
        cls,
        server_url: str,
        realm_name: str,
        client_id: str,
        username: str,
        password: str,
        authentication_realm_name: str | None = None,
        session: requests.Session | None = None,
    ):
        openid_connection = UsernamePasswordConnection(
            server_url=server_url,
            realm_name=authentication_realm_name or realm_name,
            client_id=client_id,
            username=username,
            password=password,
            session=session,
        )

        return cls.create(
            openid_connection,
            realm_name=realm_name,
        )



import os


MASTER_REALM = "master"
ADMIN_CLI_CLIENT = "admin-cli"


TEST_SERVER_URL = os.environ.get("TEST_SERVER_URL", "http://localhost:9090")

TEST_MASTER_USER = "admin"
TEST_MASTER_PASSWORD = os.environ.get("TEST_MASTER_PASSWORD", "admin")

TEST_REALM = os.environ.get("TEST_REALM", "orwell")

TEST_CLIENT_ID = os.environ.get("TEST_CLIENT_ID", "weir")
TEST_CLIENT_SECRET = os.environ.get("TEST_CLIENT_SECRET", "rockyyyy!")

TEST_USER = os.environ.get("TEST_USER", "constant")
TEST_PASSWORD = os.environ.get("TEST_PASSWORD", "rabbit")



import pytest
from . import constants
from mantelo import KeycloakAdmin
from mantelo.client import BearerAuth
from mantelo.exceptions import HttpException
import requests


@pytest.mark.integration
def test_init(openid_connection_password):
    session = requests.Session()
    adm = KeycloakAdmin(
        server_url=constants.TEST_SERVER_URL,
        realm_name=constants.TEST_REALM,
        auth=BearerAuth(openid_connection_password.token),
        session=session,
    )

    assert adm.session == session
    assert adm.realm_name == constants.TEST_REALM
    assert (
        adm.base_url
        == f"{constants.TEST_SERVER_URL}/admin/realms/{constants.TEST_REALM}"
    )

    resp = [u["username"] for u in adm.users.get()]
    assert constants.TEST_USER in resp


@pytest.mark.integration
def test_create(openid_connection_sa):
    adm = KeycloakAdmin.create(openid_connection_sa)

    assert adm.session == openid_connection_sa.session
    assert adm.realm_name == openid_connection_sa.realm_name
    assert (
        adm.base_url
        == f"{openid_connection_sa.server_url}/admin/realms/{openid_connection_sa.realm_name}"
    )
    resp = [u["username"] for u in adm.users.get()]
    assert constants.TEST_USER in resp


@pytest.mark.integration
@pytest.mark.parametrize("with_custom_session", [True, False])
def test_password_connection(with_custom_session):
    session = requests.Session() if with_custom_session else None

    adm = KeycloakAdmin.from_username_password(
        server_url=constants.TEST_SERVER_URL,
        realm_name=constants.TEST_REALM,
        client_id=constants.ADMIN_CLI_CLIENT,
        username=constants.TEST_USER,
        password=constants.TEST_PASSWORD,
        session=session,
    )

    assert adm.session == adm.session.auth.token_getter.__self__.session
    if session:
        assert adm.session == session

    resp = [u["username"] for u in adm.users.get()]
    assert constants.TEST_USER in resp


@pytest.mark.integration
def test_client_connection():
    adm = KeycloakAdmin.from_client_credentials(
        server_url=constants.TEST_SERVER_URL,
        realm_name=constants.TEST_REALM,
        client_id=constants.TEST_CLIENT_ID,
        client_secret=constants.TEST_CLIENT_SECRET,
    )

    resp = [u["username"] for u in adm.users.get()]
    assert constants.TEST_USER in resp


@pytest.mark.integration
def test_different_auth_realm(openid_connection_admin):
    adm = KeycloakAdmin.from_username_password(
        server_url=openid_connection_admin.server_url,
        realm_name=constants.TEST_REALM,
        client_id=openid_connection_admin.client_id,
        username=openid_connection_admin.username,
        password=openid_connection_admin.password,
        authentication_realm_name=openid_connection_admin.realm_name,
    )

    resp = [u["username"] for u in adm.users.get()]
    assert constants.TEST_USER in resp


@pytest.mark.integration
def test_switch_realm(openid_connection_admin):
    adm = KeycloakAdmin.create(connection=openid_connection_admin)
    assert openid_connection_admin.realm_name == constants.MASTER_REALM
    assert adm.realm_name == constants.MASTER_REALM

    resp = [u["username"] for u in adm.users.get()]
    assert constants.TEST_MASTER_USER in resp

    # Switch to another realm
    adm.realm_name = constants.TEST_REALM

    assert adm.realm_name == constants.TEST_REALM
    assert openid_connection_admin.realm_name == constants.MASTER_REALM

    resp = [u["username"] for u in adm.users.get()]
    assert constants.TEST_USER in resp


@pytest.mark.parametrize(
    ("status", "op"),
    [
        (403, lambda c: c.clients.get()),
        (400, lambda c: c.users.post({"foo": "bar"})),
        (400, lambda c: c.users.post({})),
        (405, lambda c: c.users.put({})),
        (404, lambda c: c.users("not-exist").get()),
        (
            401,
            lambda c: (
                setattr(c._store["session"], "auth", None),
                c.users.get(),
            )[-1],
        ),
    ],
)
@pytest.mark.integration
def test_exceptions(openid_connection_password, status, op):
    adm = KeycloakAdmin.create(
        connection=openid_connection_password,
        realm_name=constants.TEST_REALM,
    )

    with pytest.raises(HttpException) as excinfo:
        op(adm)

    ex = excinfo.value
    assert ex.status_code == status
    assert ex.url.startswith(constants.TEST_SERVER_URL)
    assert ex.json and isinstance(ex.json, dict)
    assert isinstance(ex.response, requests.Response)


@pytest.mark.integration
def test_headers(openid_connection_password):
    headers = {"foo": "foo", "bar": "bar"}

    adm = KeycloakAdmin.create(
        connection=openid_connection_password,
    )
    adm.session.headers.update(headers)

    # Trigger an exception to get the request object
    with pytest.raises(HttpException) as excinfo:
        adm.non_existant.get()

    hs = excinfo.value.response.request.headers
    for k, v in headers.items():
        assert hs.get(k) == v


def test_resource_private(openid_connection_password):
    adm = KeycloakAdmin.create(connection=openid_connection_password)

    with pytest.raises(AttributeError, match="_users"):
        adm._users.get()



from .version import __version__  # noqa: F401


__author__ = "Lucy Linder"
__email__ = "lucy.derlin@gmail.com"

from .client import KeycloakAdmin
from .exceptions import AuthenticationException


__all__ = ["KeycloakAdmin", "AuthenticationException"]
